{-# OPTIONS -fglasgow-exts #-}
module Conversions where

-- Metaprogramming Project
-- http://www.infosun.fmi.uni-passau.de/cl/metaprog
-- Author: Christoph A. Herrmann 
-- Copyright (C) 2006 University of Passau

import Monad
import Language.Haskell.TH
import Value
import Signal
import Indexing
import Designs
import SelZip

-- nums2words n: generates a Haskell function which converts a list of numbers
-- into an "n"-tuple of signals, taking the "n" least significant bits
nums2words :: Int -> Q Exp
nums2words n = [| $(unZipN n) . map $(num2word n) |]

-- words2nums n: generates a Haskell function which converts an
-- "n"-tuple of signals into a list of numbers
words2nums :: Int -> Q Exp
words2nums n = [| map $(word2num n) . $(zipN n) |]

-- num2word: auxiliary function for "nums2words"
num2word :: Int -> Q Exp
num2word n =
 [| \x -> case x of
           Nothing -> $(tupE (replicate n [| U |]))
           Just v  -> $(tupE [ let pow=(2::Int)^i
                               in [| if odd (v `div` pow) then I else O |]
                             | i<-[0..n-1] ])
 |]

-- word2num: auxiliary function for "nums2words"
word2num :: Int -> Q Exp
word2num n =
 do
  xs <- mapM newName [ "x"++show i | i<-[0..n-1] ]
  lamE [tupP (map varP xs)]
       (foldr (\i s -> [| case ($s,$(varE (xs!!i))) of
                           (Just s,O) -> Just (2*s)
                           (Just s,I) -> Just (2*s+1)
                           _          -> Nothing
                       |])
              [| Just 0 |]
              [0..n-1])
