{-# OPTIONS -fglasgow-exts #-}
module Designs where

-- Metaprogramming Project
-- http://www.infosun.fmi.uni-passau.de/cl/metaprog
-- Author: Christoph A. Herrmann 
-- Copyright (C) 2006 University of Passau

import List
import Language.Haskell.TH
import Signal
import CircDefs
import Skeletons
import Indexing

multiplex :: Int -> Q Exp
multiplex n =
 [| \ ~(sel,as,bs)
      -> $(skelPar (n,2,1,
                    ( \_ [[a],[b]] -> [| mux(sel,$a,$b) |])))
         (as,bs) |]

$(defFullAdder)

wordInc, wordAdder :: Int -> Q Exp
wordInc   n = skelSeq (n,1,1,\_ [[c],[x]]     -> [| halfAdder($c,$x)    |])
wordAdder n = skelSeq (n,2,1,\_ [[c],[x],[y]] -> [| fullAdder($c,$x,$y) |])

register :: Int -> Q Exp
register n = [| \ ~(write,xs) ->
                     $(let ff _ [[dat]] = [| flipflop(write,$dat) |]
                       in skelPar (n,1,1,ff)) xs |]

$(defShRegElem)

shRegister :: Int -> Q Exp
shRegister n =
  [| \ ~(write,shift,sin,xs)
          -> $(skelSeq (n,1,1,\_ [[l],[i]]
                               -> [| shRegElem (write,shift,$l,$i) |]))
             (sin,xs)
  |]

mult1 :: Int -> Q Exp
mult1 n = [| \ ~(x,ys) -> $(skelPar (n,1,1,(\_ [[yi]] ->
                                                [| and2(x,$yi) |]))) ys |]
combMult :: (Int,Int) -> Q Exp
combMult (m,n) | m>0 && n>0 =
  [| let adderN  = $(wordAdder (n-1))
         mult1N  = $(mult1 n)
         wSplitN = $(wSplit (1,(n-1)))
     in $(combMultRec (m,n) [ [|adderN|],[|mult1N|],[|wSplitN|] ]) |]

combMultRec :: (Int,Int) -> [Q Exp] -> Q Exp
combMultRec (m,n) defs@[adderN,mult1N,wSplitN]
  | m==1 = [| \ ~(op1,op2) -> $(wConcat [n,1]) ($mult1N (op1,op2), cO) |]
  | m>1  = [| \ ~(op1,op2)
                -> let  ~(x,xs)  = $(wSplit (1,m-1)) op1
                        ys       = $mult1N (x,op2)
                        zs       = $(combMultRec (m-1,n) defs) (xs,op2)
                        ~(p,ps)  = $wSplitN ys
                        ~(qs,rs) = $(wSplit (n-1,m)) zs
                        ~(cy,su) = $adderN (cO,ps,qs)
                        ~(_,inc) = $(wordInc m) (cy,rs)
                        result   = $(wConcat [1,n-1,m]) (p,su,inc)
                   in result |]

seqMult :: Bool -> (Int,Int) -> Q Exp
seqMult debug (m,n) | m>0 && n>0 =
 [| \ ~(step,init,op1,op2)
      -> let ~(xLSB,_)   = $(shRegister m)
                             (init,step,cO,$(wRev m) op1)
             yReg        = $(register n) (init,op2)
             maskedY     = $(mult1 n) (xLSB,yReg)
             newZ        = $(mult1 (m+n)) (not1(init),tmpRes)
             zReg        = $(register (m+n)) (step,newZ)
             ~(zL,zH)    = $(wSplit (m,n)) zReg
             ~(cy,sumN)  = $(wordAdder n) (cO,maskedY,zH)
             tmpRes      = $(skelWire ([m,n,1],\ [_:l,h,c]->[l++h++c]))
                           (zL,sumN,cy)
             initCount   = $(skelWire ([2],
                              \[[e,z]] -> [ e : replicate (m-1) z ]))
                           (cI,cO)
             ~(stop,_)   = $(shRegister m) (init,step,cO,initCount)
             res         = $(register (m+n)) (and2(step,stop),tmpRes)
             ready       = flipflop(step,stop)
         in $(if debug then [| (ready,res,tmpRes) |]
                       else [| (ready,res)        |])
 |]
