module Indexing where

-- Metaprogramming Project
-- http://www.infosun.fmi.uni-passau.de/cl/metaprog
-- Author: Christoph A. Herrmann 
-- Copyright (C) 2006 University of Passau

import Language.Haskell.TH
import Signal
import Skeletons

type IPLL  = [[(Int,Int)]]
type IPLLF = IPLL->IPLL

compWires :: ([Int],IPLLF) -> [[(Int,Int)]]
compWires (noInps,flow) = let inPairs = [ [ (i,j) | j<-[0..noInps!!i -1] ]
                                        | i<-[0..length noInps -1] ]
                          in flow inPairs

skelWire :: ([Int],IPLLF) -> Q Exp
skelWire (noInps,flow) = skelWireCore (noInps, compWires (noInps,flow))

rConcat, rRev, rRotUp, rRotDown :: IPLLF

rConcat  xss        = [concat xss]

rRev     [xs]       = [reverse xs]

rRotUp   [[],[x]]   = [[],[x]]
rRotUp   [xs,[x]]   = [x : init xs, [last xs]]

rRotDown [[],[y]]   = [[],[y]]
rRotDown [x:xs,[y]] = [xs ++ [y], [x]]

rDup, rSplit :: Int -> IPLLF
rDup   m [[x]] = [replicate m x]
rSplit m [xss] = [take m xss, drop m xss]

rSel :: [Int] -> IPLLF
rSel is [xs] = [[ xs!!i | i<-is ]]

wDup, wRev, wRotUp, wRotDown :: Int -> Q Exp
wDup     m = skelWire ([1],   rDup m)
wRev     m = skelWire ([m],   rRev)
wRotUp   m = skelWire ([m,1], rRotUp)
wRotDown m = skelWire ([m,1], rRotDown)

wSplit :: (Int,Int) -> Q Exp
wSplit (m,n) = skelWire ([m+n], rSplit m)

wConcat :: [Int] -> Q Exp
wConcat ms = skelWire (ms, rConcat)

wSel :: (Int,[Int]) -> Q Exp
wSel (n,inds) = skelWire ([n], rSel inds)
