{-# OPTIONS -fglasgow-exts #-}
module SelZip where

-- Metaprogramming Project
-- http://www.infosun.fmi.uni-passau.de/cl/metaprog
-- Author: Christoph A. Herrmann 
-- based on sources found on the web
-- University of Passau, 2006

import Monad
import Language.Haskell.TH


-- sel (i,n): generates a Haskell function which selects element with index "i"
-- from a tuple with "n" elements
sel :: (Int,Int) -> Q Exp
sel (i,n) = do
             xs <- sequence [ newName ("x"++show j) | j<-[0..n-1] ]
             lamE [tupP (map varP xs)] (varE (xs!!i))



-- unZipN n: generates a Haskell function, which forks a list of "n"-tuples
-- into an "n"-tuple of lists; generalization of "unzip"
unZipN :: Int -> Q Exp
unZipN n =
 [| let f []     = $(tupE (replicate n [| [] |]))
        f (x:xs) = let rs = f xs
                   in $(tupE [ [| $(sel (i,n)) x : $(sel (i,n)) rs |]
                             | i<-[0..n-1] ])
    in f
 |]




-- zipN n: generates a Haskell function, which joins an "n"-tuple of lists
-- into a list of "n"-tuples; generalization of "zip"
zipN :: Int -> Q Exp
zipN n = zipWithN n "id"

-- zipWithN n f : generates a Haskell function, which joins an "n"-tuple
-- of lists into a list of "n"-tuples, and then maps function "f" to each
-- list element; except from the uncurried form of "f" it is a generalization
-- of "zipWith"
zipWithN :: Int -> String -> Q Exp
zipWithN n f = [| let zp = $(mkZipWith n (varE (mkName f)) [|zp|]) in zp |]

-- mkZipWith: auxiliary function for "zipWithN"
mkZipWith :: Int -> Q Exp -> Q Exp -> Q Exp
mkZipWith n qf qName =
  do
   f <- qf
   name <- qName
   (pXs,eXs)   <- genPE "x"  n
   (pYs,eYs)   <- genPE "y"  n
   (pXSs,eXSs) <- genPE "xs" n
   m1 <- do
          let pcons x xs = return $ InfixP x '(:) xs
          pat  <- tupP (zipWith pcons pXs pXSs)
          body <- normalB [| $(return (AppE f (TupE eXs))) :
                             $(return (AppE name (TupE eXSs))) |]
          return $ Match pat body []
   m2 <- do
          body <- normalB [| [] |]
          pat <- tupP (take n (repeat wildP))
          return $ Match pat body []
   return $ LamE [TupP pYs] (CaseE (TupE eYs) [m1,m2])


-- genPE s n: generates "n" names with prefix "s" and their index
-- and delivers them in the form of pattern variables and expression variables
genPE :: String -> Int -> Q ([Pat],[Exp])
genPE s n = do
             let ns = [ mkName (s ++ show i) | i<-[1..n]]
             ps <- mapM varP ns
             es <- mapM varE ns
             return (ps,es)

