module Skeletons where

-- Metaprogramming Project
-- http://www.infosun.fmi.uni-passau.de/cl/metaprog
-- Author: Christoph A. Herrmann 
-- Copyright (C) 2006 University of Passau

import Language.Haskell.TH

data ConstrDesc = ConD { noNodes    :: Int,
                         nameNodes  :: Int -> Maybe String,
                         noElems    :: Int,
                         elemFun    :: Int -> [[Q Exp]] -> Q Exp,
                         elemInps   :: Int -> [[Int]],
                         elemOuts   :: Int -> [[Int]],
                         systemInps :: [[Int]],
                         systemOuts :: [[Int]]
                       }

buildSystem :: ConstrDesc -> Q Exp
buildSystem c =
  do
   (pVs,eVs) <- do
                 ns <- sequence [ newName (case nameNodes c i of
                                            Just  s -> s
                                            Nothing -> "x" ++ show i)
                                | i<-[0..noNodes c -1]]
                 ps <- mapM varP ns
                 es <- mapM varE ns
                 return (ps,es)
   decs <- sequence
             [ do
                let args = [ [ return (eVs!!k) | k<-inds ]
                           | inds <- elemInps c i ]
                rhs <- elemFun c i args
                let lhs = TildeP (TupP [ TildeP (TupP [ pVs!!k | k<-inds ])
                                       | inds <- elemOuts c i ])
                return (ValD lhs (NormalB rhs) [])
             | i <- [0..noElems c -1] ]
   let inpGrps = [ TildeP (TupP [ pVs!!ind | ind<-g ]) | g <- systemInps c ]
       outGrps = [         TupE [ eVs!!ind | ind<-g ]  | g <- systemOuts c ]
   return $ LamE [ TildeP (TupP inpGrps) ]
                 ((if null decs then id else LetE decs) (TupE outGrps))

skelWireCore :: ([Int],[[(Int,Int)]]) -> Q Exp
skelWireCore (noInps,wires) =
  buildSystem
           ConD { noNodes    = sum noInps,
                  nameNodes  = const Nothing,
                  noElems    = 0,
                  systemInps = [ take (noInps!!i) [ sum (take i noInps) ..]
                               | i<-[0..length noInps -1] ],
                  systemOuts = [ [ sum (take x noInps) + y | (x,y) <- outP ]
                               | outP <- wires ]                             
                }

skelPar :: (Int,Int,Int,(Int -> [[Q Exp]] -> Q Exp)) -> Q Exp
skelPar (n,arI,arO,f) =
  let [is,os] = [ [0..r-1] | r<-[arI,arO] ]
  in buildSystem
           ConD { noNodes    = n*(arI+arO),
                  nameNodes  = const Nothing,
                  noElems    = n,
                  elemFun    = f,
                  elemInps   = \j -> [  [      i*n+j ] | i<-is ],
                  elemOuts   = \j -> [  [(arI+o)*n+j ] | o<-os ],
                  systemInps = [ take n [      i*n ..] | i<-is ],
                  systemOuts = [ take n [(arI+o)*n ..] | o<-os ] 
                }

skelSeq :: (Int,Int,Int,(Int -> [[Q Exp]] -> Q Exp)) -> Q Exp
skelSeq (n,arI,arO,f) =
  let [is,os,ns]   = [ [0..r-1] | r<-[arI,arO,n] ]
      [xss,yss]    = [ [ p ++ show k ++ "_" ++ show j | k<-[0..m-1], j<-ns ]
                     | (p,m) <- [("x",arI),("y",arO)] ]
      as           = [ "a" ++ show j | j<-[0..n] ]
      x i j        =         i*n + j
      y o j        = (arI+  o)*n + j
      a   j        = (arI+arO)*n + j
  in buildSystem
      ConD { noNodes    = (arI+arO+1)*n+1,
             nameNodes  = \i -> Just ((xss ++ yss ++ as) !! i),
             noElems    = n,
             elemFun    = f,
             elemInps   = \j -> [a j]     : [ [x i j] | i<-is ],
             elemOuts   = \j -> [a (j+1)] : [ [y o j] | o<-os ],
             systemInps = [a 0] : [ [ x i j | j<-ns ] | i<-is ],
             systemOuts = [a n] : [ [ y o j | j<-ns ] | o<-os ]
           }
