module Value where

-- Metaprogramming Project
-- http://www.infosun.fmi.uni-passau.de/cl/metaprog
-- Author: Christoph A. Herrmann 
-- Copyright (C) 2006 University of Passau

import Language.Haskell.TH

data Value  = O | I | U  deriving (Eq,Show,Bounded,Enum)

snot1 :: Value -> Value
snot1 x = case x of { O->I; I->O; U->U }

sand2 :: (Value,Value) -> Value
sand2 xy = case xy of { (I,I)->I; (O,_)->O; (_,O)->O; _ -> U }

snand2, sor2, snor2, sxor2, seqv2 :: (Value,Value) -> Value
snand2 (a,b) = snot1 (sand2 (a,b))
sor2   (a,b) = snand2 (snot1 a, snot1 b)
snor2  (a,b) = snot1 (sor2 (a,b))
sxor2  (a,b) = sor2 (sand2 (a, snot1 b), sand2 (snot1 a, b))
seqv2  (a,b) = snot1 (sxor2 (a,b))
smux :: (Value,Value,Value) -> Value
smux (c,a,b) = sor2 (sand2 (snot1 c, a), sand2 (c, b))

-- combs n: generates all combinations of "n" values
combs :: Int -> [[Value]]
combs 0       = [[]]
combs n | n>0 = let vss = combs (n-1)
                in [ v:vs | v<-[minBound::Value .. maxBound::Value], vs<-vss ]

-- genValFun args f name: generates the declaration of a function with
-- name "name", which implements the function "f" with "args" arguments
-- by a pattern match over all combinations of arguments
genValFun :: Int -> ([Value]->Value) -> Name -> Q Dec
genValFun args f name = do
                         let x = mkName "x"
                             pcon v = ConP (mkName (show v)) []
                             econ v = ConE (mkName (show v))
                             cases = [ Match (TupP (map pcon vs))
                                             (NormalB (econ (f vs)))
                                             []
                                     | vs <- combs args ]
                         return $ FunD name
                                        [Clause [VarP x]
                                                (NormalB (CaseE (VarE x)
                                                cases)) []]
